@@ -24,7 +24,7 @@
  * your choice, without any of the additional requirements listed in clause 6
  * of the GNU Library General Public License.
  * By "a publicly distributed version of the Library", we mean either the unmodified
- * Library as distributed by INRIA, or a modified version of the Library that is
+ * Library as distributed by The Savonet Team, or a modified version of the Library that is
  * distributed under the conditions defined in clause 3 of the GNU Library General
  * Public License. This exception does not however invalidate any other reasons why
  * the executable file might be covered by the GNU Library General Public License.
@@ -181,6 +181,32 @@ CAMLprim value ocaml_ogg_sync_reset(value oy)
   CAMLreturn(Val_unit);
 }
 
+CAMLprim value ocaml_ogg_sync_pageseek(value callback, value oy)
+{
+  CAMLparam2(callback,oy);
+  CAMLlocal3(ret,s,bytes);
+  ogg_sync_state *sync = Sync_state_val(oy);
+  int len = 4096;
+  ogg_page page;
+  int err = ogg_sync_pageseek(sync, &page);
+
+  while (err <= 0)
+  {
+    ret = caml_callback(callback,Val_int(len));
+    s = Field(ret,0);
+    bytes = Field(ret,1);
+    if (Int_val(bytes) == 0)
+      caml_raise_constant(*caml_named_value("ogg_exn_eos"));
+    
+    char *buffer = ogg_sync_buffer(sync,Int_val(bytes));
+    memcpy(buffer,String_val(s),Int_val(bytes));
+    ogg_sync_wrote(sync,Int_val(bytes));
+    err = ogg_sync_pageseek(sync, &page);
+  }
+
+  CAMLreturn(value_of_page(&page));
+}
+
 CAMLprim value ocaml_ogg_sync_read(value callback, value oy)
 {
   CAMLparam2(callback,oy);
@@ -188,18 +214,22 @@ CAMLprim value ocaml_ogg_sync_read(value callback, value oy)
   ogg_sync_state *sync = Sync_state_val(oy);
   int len = 4096;
   ogg_page page;
+  int ans = ogg_sync_pageout(sync, &page);
 
-  while (ogg_sync_pageout(sync, &page) != 1)
+  while (ans != 1)
   {
+    if (ans == -1)
+      caml_raise_constant(*caml_named_value("ogg_exn_out_of_sync"));
     ret = caml_callback(callback,Val_int(len));
     s = Field(ret,0);
     bytes = Field(ret,1);
     if (Int_val(bytes) == 0)
-      caml_raise_constant(*caml_named_value("ogg_exn_not_enough_data"));
+      caml_raise_constant(*caml_named_value("ogg_exn_eos"));
 
     char *buffer = ogg_sync_buffer(sync,Int_val(bytes));
     memcpy(buffer,String_val(s),Int_val(bytes));
     ogg_sync_wrote(sync,Int_val(bytes));
+    ans = ogg_sync_pageout(sync, &page);
   }
 
   CAMLreturn(value_of_page(&page));
@@ -234,7 +264,7 @@ static inline ogg_packet *copy_packet(ogg_packet *op)
 {
   ogg_packet *nop = malloc(sizeof(ogg_packet));
   if (nop == NULL)
-    caml_failwith("malloc");
+    caml_raise_out_of_memory();
   nop->packet = malloc(op->bytes);
   memcpy(nop->packet,op->packet,op->bytes);
   nop->bytes = op->bytes;
@@ -320,9 +350,9 @@ CAMLprim value ocaml_ogg_stream_packetin(value o_stream_state, value packet)
   CAMLreturn(Val_unit);
 }
 
-CAMLprim value ocaml_ogg_stream_packetout(value callback,value o_stream_state)
+CAMLprim value ocaml_ogg_stream_packetout(value o_stream_state)
 {
-  CAMLparam2(callback,o_stream_state);
+  CAMLparam1(o_stream_state);
   ogg_stream_state *os = Stream_state_val(o_stream_state);
   ogg_packet op;
   int ret = ogg_stream_packetout(os,&op);
@@ -330,12 +360,28 @@ CAMLprim value ocaml_ogg_stream_packetout(value callback,value o_stream_state)
   if (ret == 0) 
     caml_raise_constant(*caml_named_value("ogg_exn_not_enough_data"));
 
-  if (ret == -1)
-    caml_callback(callback,Val_unit);
-
+  if (ret == -1) 
+    caml_raise_constant(*caml_named_value("ogg_exn_out_of_sync"));
+  
   CAMLreturn(value_of_packet(&op));
 }
 
+CAMLprim value ocaml_ogg_stream_packet_advance(value o_stream_state)
+{
+  CAMLparam1(o_stream_state);
+  ogg_stream_state *os = Stream_state_val(o_stream_state);
+  ogg_packet op;
+  int ret = ogg_stream_packetout(os,&op);
+
+  if (ret == 0)
+    caml_raise_constant(*caml_named_value("ogg_exn_not_enough_data"));
+
+  if (ret == -1)
+    caml_raise_constant(*caml_named_value("ogg_exn_out_of_sync"));
+
+  CAMLreturn(Val_unit);
+}
+
 CAMLprim value ocaml_ogg_stream_packetpeek(value o_stream_state)
 {
   CAMLparam1(o_stream_state);
@@ -346,9 +392,36 @@ CAMLprim value ocaml_ogg_stream_packetpeek(value o_stream_state)
   if (ret == 0)
     caml_raise_constant(*caml_named_value("ogg_exn_not_enough_data"));
 
+  if (ret == -1)
+    caml_raise_constant(*caml_named_value("ogg_exn_out_of_sync"));
+
   CAMLreturn(value_of_packet(&op));
 }
 
+CAMLprim value ocaml_ogg_stream_granulepospeek(value o_stream_state)
+{
+  CAMLparam1(o_stream_state);
+  ogg_stream_state *os = Stream_state_val(o_stream_state);
+  ogg_packet op;
+  int ret = ogg_stream_packetpeek(os,&op);
+
+  if (ret == 0)
+    caml_raise_constant(*caml_named_value("ogg_exn_not_enough_data"));
+
+  if (ret == -1)
+    caml_raise_constant(*caml_named_value("ogg_exn_out_of_sync"));
+
+  CAMLreturn(caml_copy_int64(op.granulepos));
+}
+
+CAMLprim value ocaml_ogg_stream_packet_granulepos(value _op)
+{
+  CAMLparam1(_op);
+  ogg_packet *op = Packet_val(_op);
+
+  CAMLreturn(caml_copy_int64(op->granulepos));
+}
+
 CAMLprim value ocaml_ogg_flush_stream(value o_stream_state)
 {
   CAMLparam1(o_stream_state);
@@ -415,7 +488,7 @@ CAMLprim value ocaml_ogg_skeleton_fishead(value pres_num, value pres_den, value
 
   op.packet = malloc (64);
   if (op.packet == NULL)
-    caml_failwith("malloc");
+    caml_raise_out_of_memory();
 
   memset (op.packet, 0, 64);
   memcpy (op.packet, FISHEAD_IDENTIFIER, 8); /* identifier */
